// Primitives.cpp : Defines the entry point for the console application.

#include <string>
#include <vector>
#include <algorithm>
#include <chrono>
#include "Common.h"

// OpenCL C API
#include <CL/opencl.h>

// OpenCL C++ API
#include "cl.hpp"

void task1()
{
    std::cout << "----------- Task1 begin ----------- " << std::endl;

    // Get platform
	cl_platform_id platformID;
	clGetPlatformIDs(1, &platformID, NULL);
	
    // Get Device
	cl_device_id deviceID;
	clGetDeviceIDs(platformID, CL_DEVICE_TYPE_GPU, 1, &deviceID, NULL);
	
    // Creating a context
	cl_context context;
	cl_context_properties contextProperties[] =
	{ CL_CONTEXT_PLATFORM, (cl_context_properties)platformID, 0 };
	context = clCreateContext(contextProperties, 1, &deviceID, NULL, NULL, NULL);


    // Find the maximal allocatable memory size
    cl_ulong maxSize = 0;
    clGetDeviceInfo(deviceID, CL_DEVICE_MAX_MEM_ALLOC_SIZE, sizeof(cl_ulong), &maxSize, NULL);
    size_t dataSize = static_cast<size_t>(maxSize) / sizeof(float);
	
    // Create a command queue
	cl_command_queue queue;
	queue = clCreateCommandQueue(context, deviceID, CL_QUEUE_PROFILING_ENABLE, NULL);
	
    // Creating the program
	std::string source = FileToString("../kernels/programs.cl");
	const char* csource = source.c_str();
	cl_program program = clCreateProgramWithSource(context, 1, &csource, NULL, NULL);
	cl_int err = clBuildProgram(program, 1, &deviceID, NULL, NULL, NULL);
	if (err != CL_SUCCESS)
	{
		cl_uint logLength;
		clGetProgramBuildInfo(program, deviceID, CL_PROGRAM_BUILD_LOG, 0, NULL, &logLength);
		char* log = new char[logLength];
		clGetProgramBuildInfo(program, deviceID, CL_PROGRAM_BUILD_LOG, logLength, log, 0);
		std::cout << log << std::endl;
		delete[] log;
		exit(-1);
	}

	cl_kernel kernel = clCreateKernel(program, "reduce_global", &err);
	if(!CheckCLError(err)) exit(-1);

    // Finding the largest work-group size - it is 1024
    size_t maxWorkGroupSize = 0;
    clGetKernelWorkGroupInfo(kernel, deviceID, CL_KERNEL_WORK_GROUP_SIZE, sizeof(size_t), &maxWorkGroupSize, NULL);

    // Allocating host buffer
	float* hostBuffer = new float[dataSize];
	for (size_t index = 0; index < dataSize; ++index)
	{
		hostBuffer[index] = static_cast<float>(index);
	}

    // Creating device buffer for the input
	cl_mem inputBuffer;
	inputBuffer = clCreateBuffer(context, CL_MEM_READ_ONLY, sizeof(float) * dataSize, NULL, &err);
	if (!CheckCLError(err)) exit(-1);

	clEnqueueWriteBuffer(queue, inputBuffer, CL_TRUE, 0, sizeof(float) * dataSize, hostBuffer, 0, NULL, NULL);

    // Creating device buffer for the output
	cl_mem outputBuffer;
    size_t outputSize = (dataSize / 2 + maxWorkGroupSize - 1) / maxWorkGroupSize; //number of workgroups
	outputBuffer = clCreateBuffer(context, CL_MEM_WRITE_ONLY, sizeof(float) * outputSize, NULL, &err);
	if (!CheckCLError(err)) exit(-1);

    unsigned int kernelNum = dataSize / 2;

    cl_mem* input = &inputBuffer;
    cl_mem* output = &outputBuffer;

    std::chrono::high_resolution_clock::time_point gpuT1 = std::chrono::high_resolution_clock::now();

    //TODO Call the kernel function multiple times until the largest number in the array is found
    //put the result in the gpuMaxValue variable












    float gpuMaxValue = 0.0f;

    std::chrono::high_resolution_clock::time_point gpuT2 = std::chrono::high_resolution_clock::now();

    std::chrono::high_resolution_clock::time_point cpuT1 = std::chrono::high_resolution_clock::now();

    float cpuMaxValue = 0.0f;

    for (unsigned int i = 0; i < dataSize; i++)
    {
        if (hostBuffer[i] > cpuMaxValue)
        {
            cpuMaxValue = hostBuffer[i];
        }
    }

    std::chrono::high_resolution_clock::time_point cpuT2 = std::chrono::high_resolution_clock::now();

	std::cout << "Result GPU: "  << gpuMaxValue << std::endl;
    std::cout << "Execution time GPU: " << std::chrono::duration_cast<std::chrono::microseconds>(gpuT2 - gpuT1).count() << "us" << std::endl;

    std::cout << "Result CPU: " << cpuMaxValue << std::endl;
    std::cout << "Execution time CPU: " << std::chrono::duration_cast<std::chrono::microseconds>(cpuT2 - cpuT1).count() << "us" << std::endl;

    // Releasing stuff
    delete[] hostBuffer;

    clReleaseMemObject(inputBuffer);
    clReleaseMemObject(outputBuffer);
    clReleaseDevice(deviceID);
    clReleaseContext(context);

    std::cout << "----------- Task1 end ----------- " << std::endl;
}

void task2()
{
    std::cout << "----------- Task2 begin ----------- " << std::endl;

    // Get platform
    cl_platform_id platformID;
    clGetPlatformIDs(1, &platformID, NULL);

    // Get device
    cl_device_id deviceID;
    clGetDeviceIDs(platformID, CL_DEVICE_TYPE_GPU, 1, &deviceID, NULL);

    // Creating a context
    cl_context context;
    cl_context_properties contextProperties[] =
    { CL_CONTEXT_PLATFORM, (cl_context_properties)platformID, 0 };
    context = clCreateContext(contextProperties, 1, &deviceID, NULL, NULL, NULL);

    // Finding the size of the device memory
    cl_ulong deviceMemory = 0;
    clGetDeviceInfo(deviceID, CL_DEVICE_GLOBAL_MEM_SIZE, sizeof(cl_ulong), &deviceMemory, NULL);
    deviceMemory /= 2;

    // how many tasks we want to run
    int instances = 5;
    size_t dataSize = deviceMemory / instances / 2;

    cl_int err;

    // Creating command queues
    cl_command_queue inOrderQueue;
    cl_command_queue outOfOrderQueue;

    //TODO: create two command queues, one with in order execution model and one with out of order execution model





    // Creating the program
    std::string source = FileToString("../kernels/programs.cl");
    const char* csource = source.c_str();
    cl_program program = clCreateProgramWithSource(context, 1, &csource, NULL, NULL);
    err = clBuildProgram(program, 1, &deviceID, NULL, NULL, NULL);
    if (err != CL_SUCCESS)
    {
        cl_uint logLength;
        clGetProgramBuildInfo(program, deviceID, CL_PROGRAM_BUILD_LOG, 0, NULL, &logLength);
        char* log = new char[logLength];
        clGetProgramBuildInfo(program, deviceID, CL_PROGRAM_BUILD_LOG, logLength, log, 0);
        std::cout << log << std::endl;
        delete[] log;
        exit(-1);
    }

    // Allocating input and output buffers
    std::vector<cl_mem> inputBuffers;
    std::vector<cl_mem> outputBuffers;

    float* hostBuffer = new float[dataSize];
    memset(hostBuffer, 0, sizeof(float) * dataSize);

    for (int i = 0; i < instances; ++i)
    {
        inputBuffers.push_back(clCreateBuffer(context, CL_MEM_READ_ONLY, sizeof(float) * dataSize, NULL, &err));
        if (!CheckCLError(err)) exit(-1);
        clEnqueueWriteBuffer(inOrderQueue, inputBuffers[i], CL_TRUE, 0, sizeof(float) * dataSize, hostBuffer, 0, NULL, NULL);

        outputBuffers.push_back(clCreateBuffer(context, CL_MEM_READ_ONLY, sizeof(float) * dataSize, NULL, &err));
        if (!CheckCLError(err)) exit(-1);
        clEnqueueWriteBuffer(inOrderQueue, outputBuffers[i], CL_TRUE, 0, sizeof(float) * dataSize, hostBuffer, 0, NULL, NULL);
    }

    cl_kernel kernel = clCreateKernel(program, "long_job", &err);
    if (!CheckCLError(err)) exit(-1);

    std::chrono::high_resolution_clock::time_point inOrderT1 = std::chrono::high_resolution_clock::now();

    //TODO: copy the host buffer to the device, run the kernel and copy back the result
    // use the command queue with in order execution model

    for (int i = 0; i < instances; i++)
    {






    }

    clFinish(inOrderQueue);

    std::chrono::high_resolution_clock::time_point inOrderT2 = std::chrono::high_resolution_clock::now();

    std::vector<cl_event> events(instances * 2);

    std::chrono::high_resolution_clock::time_point outOfOrderT1 = std::chrono::high_resolution_clock::now();

    //TODO: copy the host buffer to the device, run the kernel and copy back the result
    // use the command queue with out of order execution model
    // take care with synchronization - use events

    for (int i = 0; i < instances; i++)
    {






    }

    clFinish(outOfOrderQueue);

    std::chrono::high_resolution_clock::time_point outOfOrderT2 = std::chrono::high_resolution_clock::now();    

    std::cout << "Execution time InOrder: " << std::chrono::duration_cast<std::chrono::microseconds>(inOrderT2 - inOrderT1).count() << "us" << std::endl;

    std::cout << "Execution time OutOfOrder: " << std::chrono::duration_cast<std::chrono::microseconds>(outOfOrderT2 - outOfOrderT1).count() << "us" << std::endl;

    delete[] hostBuffer;

    for (auto b : inputBuffers)
    {
        clReleaseMemObject(b);
    }

    for (auto b : outputBuffers)
    {
        clReleaseMemObject(b);
    }

    clReleaseDevice(deviceID);
    clReleaseContext(context);

    std::cout << "----------- Task2 end ----------- " << std::endl;
}

int main()
{
	task1();

    std::cout << std::endl;

	task2();

    int tmp;
    std::cin >> tmp;

    return 0;
}

